from unittest.mock import Mock, patch
import os.path
import json

from secpy.core.ticker_company_exchange_map import TickerCompanyExchangeMap

RESOURCES = os.path.join(os.path.dirname(__file__), "..", "resources")


@patch("secpy.core.network_client.requests.get")
def mock_company_tickers_exchange(mock_get):
    mock_cte_path = os.path.join(RESOURCES, "company_tickers_exchange.json")
    with open(mock_cte_path, "r") as f:
        mock_get.return_value = Mock(ok=True)
        mock_get.return_value.json.return_value = json.load(f)
        ticker_company_exchange_map = TickerCompanyExchangeMap("/")
        ticker_company_exchange_map.list_ciks()
        return ticker_company_exchange_map




import unittest

from secpy.core.utils.cik_opts import CIKOpts


class CIKOptsTest(unittest.TestCase):
    def test_valid_cik_arg_on_valid_ciks(self):
        cik1 = "0000012345"
        cik2 = "1000012349"
        cik3 = "0021923012"
        cik4 = "0000000001"
        ciks = [cik1, cik2, cik3, cik4]
        for cik in ciks:
            self.assertTrue(CIKOpts.validate_cik_arg(cik))

    def test_valid_cik_arg_on_invalid_args(self):
        ticker = "BLH"
        non_numeric = "a000013423"
        invalid_length = "0001234"
        invalid_ciks = [ticker, non_numeric, invalid_length]
        for invalid_cik in invalid_ciks:
            self.assertRaises(AssertionError, CIKOpts.validate_cik_arg, invalid_cik)

    def test_format_cik_int(self):
        cik_int = 12345
        formatted_cik = CIKOpts.format_cik(cik_int)
        expected_cik = "0000012345"
        self.assertEqual(formatted_cik, expected_cik)

    def test_format_cik_str(self):
        cik_int = "12345"
        formatted_cik = CIKOpts.format_cik(cik_int)
        expected_cik = "0000012345"
        self.assertEqual(formatted_cik, expected_cik)


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest.mock import patch, Mock
import os
import json

from secpy.core.ticker_company_exchange_map import TickerCompanyExchangeMap, CTEObject

MOCK_CTE_DATA = os.path.join(os.path.dirname(__file__), "..", "..", "resources", "company_tickers_exchange.json")


class TickerCompanyExchangeMapTest(unittest.TestCase):

    @patch("secpy.core.network_client.requests.get")
    def test_lookup_ticker(self, mock_get):
        with open(MOCK_CTE_DATA, "r") as f:
            mock_get.return_value = Mock(ok=True)
            mock_get.return_value.json.return_value = json.load(f)
        ticker_company_exchange_map = TickerCompanyExchangeMap("/")
        actual_cte_object = ticker_company_exchange_map.lookup_ticker("MSFT")
        expected_cte_object = CTEObject([789019, "MICROSOFT CORP", "MSFT", "Nasdaq"])
        self.assertEqual(actual_cte_object, expected_cte_object)

    @patch("secpy.core.network_client.requests.get")
    def test_lookup_cik(self, mock_get):
        with open(MOCK_CTE_DATA, "r") as f:
            mock_get.return_value = Mock(ok=True)
            mock_get.return_value.json.return_value = json.load(f)

        ticker_company_exchange_map = TickerCompanyExchangeMap("/")
        actual_cte_object = ticker_company_exchange_map.lookup_cik("0000789019")
        expected_cte_object = CTEObject([789019, "MICROSOFT CORP", "MSFT", "Nasdaq"])
        self.assertEqual(actual_cte_object, expected_cte_object)


if __name__ == '__main__':
    unittest.main()



from secpy.company_facts import CompanyFactsEndpoint, CompanyFactsBulkEndpoint
from secpy.submissions import SubmissionsEndpoint, SubmissionsBulkEndpoint
from secpy.company_concept import CompanyConceptEndpoint
from secpy.frames import FramesEndpoint
from secpy.core.ticker_company_exchange_map import TickerCompanyExchangeMap


class SECPyClient:
    def __init__(self, user_agent):
        self.user_agent = user_agent

    def submissions(self, **kwargs):
        return SubmissionsEndpoint(self.user_agent, **kwargs)

    def bulk_submissions(self, existing_archive=None,  **kwargs):
        return SubmissionsBulkEndpoint(self.user_agent, existing_archive, **kwargs)

    def company_facts(self, **kwargs):
        return CompanyFactsEndpoint(self.user_agent, **kwargs)

    def bulk_company_facts(self, existing_archive=None, **kwargs):
        return CompanyFactsBulkEndpoint(self.user_agent, existing_archive, **kwargs)

    def company_concepts(self, **kwargs):
        return CompanyConceptEndpoint(self.user_agent, **kwargs)

    def frames(self, **kwargs):
        return FramesEndpoint(self.user_agent, **kwargs)

    def ticker_company_exchange_map(self, **kwargs):
        return TickerCompanyExchangeMap(self.user_agent, **kwargs)



from abc import ABC, abstractmethod
import os
import shutil
from zipfile import ZipFile
from tempfile import gettempdir
import re
import json
import time

from secpy.core.mixins.base_endpoint_mixin import BaseEndpointMixin


class BulkDataEndpoint(BaseEndpointMixin, ABC):
    def __init__(self, user_agent, existing_archive=None, **kwargs):
        """
        Base class for downloading and parsing SEC bulk ZIP files.
        ZIP files are updated everyday at approximately 3:00 AM (EST)
        @param existing_archive: str, full path to existing archive file to parse instead of downloading new ZIP file from SEC
        @param kwargs:
        """
        super().__init__(user_agent, **kwargs)
        self._archive_persisted = existing_archive is not None
        self.bulk_data_file_object = self.__set_bulk_data_file_object(existing_archive)

    def __set_bulk_data_file_object(self, existing_archive):
        return BulkDataFileObject(existing_archive, self._ticker_cte_map) if existing_archive else None

    def download_bulk_data(self, override=False, chunk_size=1028):
        """
        Download from bulk data endpoint to a temp directory and set the bulk_data_file_object
        @param override: bool, override existing_archive and download
        @param chunk_size: int, number of bytes to read into memory while iterating over response
        @return:
        """
        if not self._archive_persisted or (self._archive_persisted and override):
            archive_path = self.__get_temp_zip_file_path()
            self._validate_path_and_download_file(self._endpoint, archive_path, chunk_size=chunk_size)
            self.bulk_data_file_object = BulkDataFileObject(archive_path, self._ticker_cte_map)
        else:
            self._logger.warn("Archive already exists, skipping download...")
        return self

    @staticmethod
    def __get_temp_zip_file_path():
        temp_dir = gettempdir()
        filename = str(int(time.time()))
        return os.path.join(temp_dir, filename) + ".zip"

    def get_data_for_ticker_from_archive(self, ticker):
        """
        Parses data from the file associated w/ ticker from the zip file. Downloads the zip file if it doesn't exist.
        @param ticker: converted to CIK to find file to parse from archive
        @return:
        """
        self._get_bulk_data_if_none()
        data = self.bulk_data_file_object.get_json_for_ticker_from_zip(ticker)
        return self._parse_data(data)

    def persist_zipfile(self, output_path):
        """
        Persists the bulk data zip file to a non-temporary location specified by output_path
        @param output_path: Full path to persists the bulk data zip file
        @return: None
        """
        if self._archive_persisted:
            raise Exception("Archive at path {} has already been persisted!".format(self.bulk_data_file_object.archive_path))

        if self.bulk_data_file_object.archive_exists():
            shutil.move(self.bulk_data_file_object.archive_path, output_path)
            self.__set_bulk_data_file_object(output_path)
            self._archive_persisted = True
        else:
            raise Exception("Bulk data not loaded into memory, nothing to save!")

    def _get_bulk_data_if_none(self, **kwargs):
        if not self.bulk_data_file_object:
            self.download_bulk_data(**kwargs)

    def parsed_data_generator(self, filter_unlisted=True):
        """
        Creates generator of parsed data objects
        @return: generator of parsed data objects
        """
        file_filter = self.bulk_data_file_object.unlisted_companies if filter_unlisted else []
        self._get_bulk_data_if_none()
        return (self.__extract_file_and_parse_data(filename) for filename in self.bulk_data_file_object.get_filelist() if filename not in file_filter)

    def for_all_companies(self, func):
        """
        Applies some function to all data in archive, ie min, max, avg of some value
        @param func: python callable, function to apply to entire set of data in archive
        @return: return value of func
        """
        return func(self.__extract_file_and_parse_data(filename) for filename in self.bulk_data_file_object.get_filelist())

    def for_each_company(self, func):
        """
        Applies some function to each data object in the archive
        @param func: python callable, function to apply to each data object
        @return: list of return value of func
        """
        self._get_bulk_data_if_none()
        return [func(self.__extract_file_and_parse_data(filename)) for filename in self.bulk_data_file_object.get_filelist()]

    def __extract_file_and_parse_data(self, filename):
        data = self.bulk_data_file_object.get_file(filename)
        return self._parse_data(data)

    def __del__(self):
        """
        Deletes bulk data zip file from temporary location after object goes out of scope
        @return:
        """
        if not self._archive_persisted and self.bulk_data_file_object and self.bulk_data_file_object.archive_exists():
            os.remove(self.bulk_data_file_object.archive_path)

    @abstractmethod
    def _parse_data(self, data):
        """
        Parses response data into some subclass of BaseDataObjectMixin
        @param data: dict, data to parse
        @return: parsed BaseDataObjectMixin instance
        """
        pass


class BulkDataFileObject:
    __FILENAME_REGEX = re.compile("CIK\d{10}.json$")

    def __init__(self, archive_path, ticker_cte_map):
        """
        Handles the parsing of files in bulk data zip file and the creation of CIK/ticker to filename maps.
        @param archive_path: Full path of bulk data zip file
        @param ticker_cte_map: ticker to CTE map object used to create the CIK/Ticker to filename maps
        """
        self.archive_path = archive_path
        self.__zipfile = self.__set_zip_file(archive_path)
        self.__ticker_to_filename_map = {}
        self.__cik_to_filename_map = {}
        self.unlisted_companies = []

        self.__create_filename_maps(ticker_cte_map)

    @staticmethod
    def __set_zip_file(archive_path):
        assert os.path.exists(archive_path)
        return ZipFile(archive_path)

    def __create_filename_maps(self, ticker_cte_map):
        """
        Creates the CIK/ticker -> filename mappings from files in bulk ZIP file
        @param ticker_cte_map: ticker -> CTEObject map. Used to map CIK from filenames -> ticker
        @return: Unit
        """
        file_list = self.__zipfile.filelist
        valid_file_list = list(filter(lambda file: self.__filename_matches_pattern(file.filename), file_list))
        valid_file_names = [file.filename for file in valid_file_list]
        for filename in valid_file_names:
            cik = self.__parse_cik_from_filename(filename)
            ticker = ticker_cte_map.lookup_cik(cik)
            if ticker == ticker_cte_map.UNKNOWN:
                self.__insert_to_cik_to_filename_map(cik, filename)
                self.__insert_to_unlisted_ciks(cik)
            else:
                self.__insert_to_cik_to_filename_map(cik, filename)
                self.__insert_to_ticker_to_filename_map(ticker.ticker, filename)

    @staticmethod
    def __parse_cik_from_filename(filename):
        return "".join(re.findall('\d+', filename))

    def __filename_matches_pattern(self, filename):
        return bool(self.__FILENAME_REGEX.match(filename))

    def __insert_to_ticker_to_filename_map(self, ticker, filename):
        self.__ticker_to_filename_map[ticker] = filename

    def __insert_to_cik_to_filename_map(self, cik, filename):
        self.__cik_to_filename_map[cik] = filename

    def __insert_to_unlisted_ciks(self, cik):
        self.unlisted_companies.append(cik)

    def __get_filename_from_ticker(self, ticker):
        return self.__ticker_to_filename_map[ticker]

    def __get_filename_from_cik(self, cik):
        return self.__cik_to_filename_map[cik]

    def get_file(self, filename):
        """
        Opens file in zip archive and parses it to json
        @param filename: file to parse
        @return: dict, parsed json
        """
        return json.loads(self.__zipfile.read(filename))

    def get_filelist(self):
        """
        Get list of all files in archive
        @return: list
        """
        return self.__zipfile.filelist

    def get_json_for_ticker_from_zip(self, ticker):
        """
        Gets json data from archive for a particular ticker
        @param ticker: str, ticker to extract data for
        @return: dict
        """
        filename = self.__get_filename_from_ticker(ticker)
        return self.get_file(filename)

    def get_json_for_cik_from_zip(self, cik):
        """
        Gets json data from archive for a particular cik
        @param cik: str, cik to extract data for
        @return: dict
        """
        filename = self.__get_filename_from_cik(cik)
        return self.get_file(filename)

    def archive_exists(self):
        """
        Checks to see if the underlying zip archive still exists
        @return: bool
        """
        return os.path.exists(self.archive_path)

    def get_ticker_to_filename_map(self):
        """
        Gets mapping of ticker to filenames
        @return:
        """
        return self.__ticker_to_filename_map

    def get_cik_to_filename_map(self):
        """
        Gets mapping of cik to filenames
        @return:
        """
        return self.__cik_to_filename_map



from os import path

from secpy.core.endpoint_enum import EndpointEnum
from secpy.core.network_client import NetworkClient


class BaseNetworkClientMixin:
    def __init__(self, user_agent, **kwargs):
        """
        Mixin for for adding network_client to a class
        @param user_agent: email address to use in headers when making requests to SEC REST API
        @param kwargs:
        """
        self.__network_client = NetworkClient(user_agent, **kwargs)

    def _validate_args_and_make_request(self, endpoint, **kwargs):
        assert EndpointEnum.validate_endpoint_kwargs(**kwargs)
        return self.__network_client.make_request_json(endpoint, **kwargs)

    def _validate_path_and_download_file(self, endpoint, target_path, **kwargs):
        assert not path.exists(target_path), "target_path {} already exists!".format(target_path)
        return self.__network_client.download_file(endpoint, target_path, **kwargs)



import unittest

from secpy.core.endpoint_enum import EndpointEnum


class EndpointEnumTest(unittest.TestCase):
    def test_validate_endpoint_kwargs(self):
        kwargs = {"CIK": "0000012345", "PERIOD_FORMAT": "CY2020"}
        self.assertTrue(EndpointEnum.validate_endpoint_kwargs(**kwargs))

    def test_validate_endpoint_kwargs_invalid(self):
        kwargs = {"CIK": "0000012", "PERIOD_FORMAT": "CY2020"}
        self.assertRaises(AssertionError, EndpointEnum.validate_endpoint_kwargs, **kwargs)


if __name__ == '__main__':
    unittest.main()



from enum import Enum
from secpy.core.utils.cik_opts import CIKOpts
from secpy.core.utils.period_format_opts import PeriodFormatOpts


class EndpointEnum(Enum):
    """
    Specifies templates for all SEC REST API endpoints that are used in secpy
    """
    __BASE_SEC_URL = "https://www.sec.gov"
    __BASE_DATA_SEC_URL = "https://data.sec.gov"

    EDGAR_DATA_ARCHIVES = __BASE_SEC_URL + "/Archives/edgar/data/{CIK_NUM}/{ACCESSION_NUM}/{FILE_NAME}"
    COMPANY_TICKER_EXCHANGE = __BASE_SEC_URL + "/files/company_tickers_exchange.json"
    BULK_COMPANY_FACTS = __BASE_SEC_URL + "/Archives/edgar/daily-index/xbrl/companyfacts.zip"
    BULK_SUBMISSIONS = __BASE_SEC_URL + "/Archives/edgar/daily-index/bulkdata/submissions.zip"

    SUBMISSIONS = __BASE_DATA_SEC_URL + "/submissions/{FILE_NAME}"
    SUBMISSIONS_CIK = __BASE_DATA_SEC_URL + "/submissions/CIK{CIK}.json"

    COMPANY_CONCEPT = __BASE_DATA_SEC_URL + "/api/xbrl/companyconcept/CIK{CIK}/{TAXONOMY}/{CONCEPT}.json"
    COMPANY_FACTS = __BASE_DATA_SEC_URL + "/api/xbrl/companyfacts/CIK{CIK}.json"

    FRAMES = __BASE_DATA_SEC_URL + "/api/xbrl/frames/{TAXONOMY}/{CONCEPT}/{UNIT}/{PERIOD_FORMAT}.json"

    @staticmethod
    def validate_endpoint_kwargs(**kwargs):
        """
        Validates any verifiable kwarg arguments
        @param kwargs: key-value argument to validate
        @return: bool
        """
        endpoint_kwargs = {k: v for k, v in kwargs.items() if k in _VERIFIABLE_ARGS_MAP.keys()}
        for key, val in endpoint_kwargs.items():
            validation_func = _VERIFIABLE_ARGS_MAP[key]
            validation_func(val)
        return True


_VERIFIABLE_ARGS_MAP = {
    "CIK": CIKOpts.validate_cik_arg,
    "PERIOD_FORMAT": PeriodFormatOpts.validate_period_format_arg
}




CIK_LENGTH = 10


class CIKOpts:
    @staticmethod
    def validate_cik_arg(cik):
        """
        Validate that a given CIK is valid. A valid CIK is 10-digit numeric value
        @return: bool
        """
        assert cik.isnumeric(), "Input CIK {} is non-numeric!".format(cik)
        assert len(cik) == CIK_LENGTH, "Input CIK {} must be a {} digit number".format(cik, CIK_LENGTH)
        return True

    @staticmethod
    def format_cik(cik):
        """
        Pads given cik_num w/ additional 0s so that the result is a 10 digit numeric string
        @param cik: CIK integer to format
        @return: string, 10 digit numeric string
        """
        formatted_cik = str(cik).zfill(CIK_LENGTH)
        return formatted_cik







from enum import Enum

from secpy.core.bulk_data import BulkDataEndpoint
from secpy.core.endpoint_enum import EndpointEnum
from secpy.core.mixins.base_data_object_mixin import BaseDataObjectMixin
from secpy.core.mixins.base_endpoint_mixin import BaseEndpointMixin
from secpy.core.network_client import NetworkClient
from secpy.core.utils.cik_opts import CIKOpts


class SubmissionsEndpoint(BaseEndpointMixin):
    """
    Handles the downloading and parsing of Submissions data from the SEC REST API
    """
    _endpoint = EndpointEnum.SUBMISSIONS_CIK

    def get_submissions_for_ticker(self, ticker):
        """
        Request submissions data from SEC REST API for a given ticker. Converts ticker -> CIK and then makes request
        @param ticker: str
        @return: Submissions
        """
        cte_object = self._ticker_cte_map.lookup_ticker(ticker)
        return self.get_submission_for_cik(cte_object.cik)

    def get_submission_for_cik(self, cik):
        """
        Request submissions data from SEC REST API for a given CIK.
        @param cik: str
        @return: Submissions
        """
        response = self._validate_args_and_make_request(self._endpoint, CIK=cik)
        return Submissions(response)


class SubmissionsBulkEndpoint(BulkDataEndpoint):
    """
    Handles the downloading and parsing of the bulk Submissions zip file
    """
    _endpoint = EndpointEnum.BULK_SUBMISSIONS

    def _parse_data(self, data):
        return Submissions(data)


class Submissions(BaseDataObjectMixin):
    class SubmissionsSchemaEnum(Enum):
        CIK = "cik"
        ENTITY_TYPE = "entityType"
        SIC = "sic"
        SIC_DESCRIPTION = "sicDescription"
        INSIDER_TRANSACTION_FOR_OWNER_EXISTS = "insiderTransactionForOwnerExists"
        INSIDER_TRANSACTION_FOR_ISSUER_EXISTS = "insiderTransactionForIssuerExists"
        NAME = "name"
        TICKERS = "tickers"
        EXCHANGES = "exchanges"
        EIN = "ein"
        DESCRIPTION = "description"
        WEBSITE = "website"
        INVESTOR_WEBSITE = "investorWebsite"
        CATEGORY = "category"
        FISCAL_YEAR_END = "fiscalYearEnd"
        STATE_OF_INCORPORATION = "stateOfIncorporation"
        STATE_OF_INCORPORATION_DESCRIPTION = "stateOfIncorporationDescription"
        ADDRESSES = "addresses"
        PHONE = "phone"
        FLAGS = "flags"
        FORMER_NAMES = "formerNames"
        FILINGS = "filings"

    def __init__(self, data):
        """
        Container class for data retrieved from Submissions endpoint for a given company.
        Includes various metadata on the company and a list of filings made by the company
        @param data: dict
        """
        self.cik = CIKOpts.format_cik(data[self.SubmissionsSchemaEnum.CIK.value])
        self.entity_type = data[self.SubmissionsSchemaEnum.ENTITY_TYPE.value]
        self.sic = data[self.SubmissionsSchemaEnum.SIC.value]
        self.sic_description = data[self.SubmissionsSchemaEnum.SIC_DESCRIPTION.value]
        self.insider_transaction_for_owner_exists = data[self.SubmissionsSchemaEnum.INSIDER_TRANSACTION_FOR_OWNER_EXISTS.value]
        self.insider_transaction_for_issuer_exists = data[self.SubmissionsSchemaEnum.INSIDER_TRANSACTION_FOR_ISSUER_EXISTS.value]
        self.name = data[self.SubmissionsSchemaEnum.NAME.value]
        self.tickers = data[self.SubmissionsSchemaEnum.TICKERS.value]
        self.exchanges = data[self.SubmissionsSchemaEnum.EXCHANGES.value]
        self.ein = data[self.SubmissionsSchemaEnum.EIN.value]
        self.description = data[self.SubmissionsSchemaEnum.DESCRIPTION.value]
        self.website = data[self.SubmissionsSchemaEnum.WEBSITE.value]
        self.investor_website = data[self.SubmissionsSchemaEnum.INVESTOR_WEBSITE.value]
        self.category = data[self.SubmissionsSchemaEnum.CATEGORY.value]
        self.fiscal_year_end = data[self.SubmissionsSchemaEnum.FISCAL_YEAR_END.value]
        self.state_of_incorporation = data[self.SubmissionsSchemaEnum.STATE_OF_INCORPORATION.value]
        self.state_of_incorporation_description = data[self.SubmissionsSchemaEnum.STATE_OF_INCORPORATION_DESCRIPTION.value]
        self.addresses = self.__set_addresses(data)
        self.phone = data[self.SubmissionsSchemaEnum.PHONE.value]
        self.flags = data[self.SubmissionsSchemaEnum.FLAGS.value]
        self.former_names = data[self.SubmissionsSchemaEnum.FORMER_NAMES.value]
        self.filings = self.__set_filings(data)

    def __set_addresses(self, data):
        addresses = data[self.SubmissionsSchemaEnum.ADDRESSES.value]
        return {k: Address(v) for k, v in addresses.items()}

    def __set_filings(self, data):
        filings = data[self.SubmissionsSchemaEnum.FILINGS.value]
        return Filings(filings, self.cik)


class Address(BaseDataObjectMixin):
    class AddressesSchemaEnum(Enum):
        STREET_1 = "street1"
        STREET_2 = "street2"
        CITY = "city"
        STATE_OR_COUNTRY = "stateOrCountry"
        STATE_OR_COUNTRY_DESCRIPTION = "stateOrCountryDescription"
        ZIP_CODE = "zipCode"

    def __init__(self, data):
        """
        Represents primary address(es) of a given company based on their filings
        @param data: dict
        """
        self.street_1 = data[self.AddressesSchemaEnum.STREET_1.value]
        self.street_2 = data[self.AddressesSchemaEnum.STREET_2.value]
        self.city = data[self.AddressesSchemaEnum.CITY.value]
        self.state_or_country = data[self.AddressesSchemaEnum.STATE_OR_COUNTRY.value]
        self.state_or_country_description = data[self.AddressesSchemaEnum.STATE_OR_COUNTRY_DESCRIPTION.value]
        self.zip_code = data[self.AddressesSchemaEnum.ZIP_CODE.value]


class HasFilingsMixin(BaseDataObjectMixin):
    def __init__(self, cik):
        self.cik = cik

    def _parse_filings(self, data):
        """
        The 'recent' field returned submissions api endpoint is an object consisting of numerous arrays
        each containing a single element of an overall filing object. The fields of a filing can be stitched together by
        grabbing the i'th index of each field, putting it in a dict keyed by the fieldnames and then passing that dict to a Filing object
        @param data: dict
        @return: List[Filing]
        """
        first_key = list(data.keys())[0]
        num_filings = len(data[first_key])
        return [Filing({k: v[i] for k, v in data.items()}, self.cik) for i in range(num_filings)]


class Filings(HasFilingsMixin):
    class FilingsSchemaEnum(Enum):
        RECENT = "recent"
        FILES = "files"

    def __init__(self, data, cik):
        """
        Represents historical/recent filings for a given company
        @param data: dict
        @param cik: str
        """
        super().__init__(cik)
        self.cik = cik
        self.recent_files = self.__set_recent_filings(data)
        self.historical_files = self.__set_historical_filings(data)

    def __set_recent_filings(self, data):
        recent_filings = data[Filings.FilingsSchemaEnum.RECENT.value]
        return self._parse_filings(recent_filings)

    def __set_historical_filings(self, data):
        historical_files = data[self.FilingsSchemaEnum.FILES.value]
        return [HistoricalFiling(historical_file, self.cik) for historical_file in historical_files]

    def filter_by_form(self, form_type):
        """
        Filter available recent filings by a particular form type
        @param form_type: str
        @return: List[Filing]
        """
        return [form for form in self.recent_files if form.form == form_type]


class HistoricalFiling(HasFilingsMixin):
    class Filing(Enum):
        NAME = "name"
        FILING_COUNT = "filingCount"
        FILING_FROM = "filingFrom"
        FILING_TO = "filingTo"

    def __init__(self, data, cik):
        """
        Per the SEC REST API documentation:
            "The object’s property path contains at least one year’s of filing or to 1,000 (whichever is more) of the most
             recent filings in a compact columnar data array. If the entity has additional filings,
             files will contain an array of additional JSON files and the date range for the filings each one contains."
        This represents the array of additional JSON files that represent excess filings not represented in the initial submissions request
        Specifies the date range covered by a particular set of historical filings and allows for downloading/parsing the additional JSON files
        as Filings instances
        @param data: dict
        @param cik: str, cik associated w/ company
        """
        super().__init__(cik)
        self.cik = cik
        self.filename = data[self.Filing.NAME.value]
        self.filing_count = data[self.Filing.FILING_COUNT.value]
        self.filing_from = data[self.Filing.FILING_FROM.value]
        self.filing_to = data[self.Filing.FILING_TO.value]
        self.file_link = self.__set_file_path()

    def __set_file_path(self):
        return EndpointEnum.SUBMISSIONS.value.format(FILE_NAME=self.filename)

    def get_historical_filings(self, user_agent):
        """
        Makes request to download historical filings data from self.file_link
        @param user_agent: str, used in header of request to identify application making the request
        @return: List[Submission]
        """
        nwc = NetworkClient(user_agent)
        response = nwc.make_request_json(EndpointEnum.SUBMISSIONS, FILE_NAME=self.filename)
        return self._parse_filings(response)


class Filing:
    class FilingSchemaEnum(Enum):
        ACCESSION_NUMBER = "accessionNumber"
        FILING_DATE = "filingDate"
        REPORT_DATE = "reportDate"
        ACCEPTANCE_DATE_TIME = "acceptanceDateTime"
        ACT = "act"
        FORM = "form"
        FILE_NUMBER = "fileNumber"
        FILM_NUMBER = "filmNumber"
        ITEMS = "items"
        SIZE = "size"
        IS_XBRL = "isXBRL"
        IS_INLINE_XBRL = "isInlineXBRL"
        PRIMARY_DOCUMENT = "primaryDocument"
        PRIMARY_DOCUMENT_DESCRIPTION = "primaryDocumentDescription"

    def __init__(self, data, cik):
        """
        Represents a single filing made by a given company.
        Provides various metadata regarding the filing and means w/ which to download the filing
        @param data: dict
        @param cik: cik of the specified company
        """
        self.cik = cik
        self.accession_number = data[self.FilingSchemaEnum.ACCESSION_NUMBER.value]
        self.filing_date = data[self.FilingSchemaEnum.FILING_DATE.value]
        self.report_date = data[self.FilingSchemaEnum.REPORT_DATE.value]
        self.acceptance_date_time = data[self.FilingSchemaEnum.ACCEPTANCE_DATE_TIME.value]
        self.act = data[self.FilingSchemaEnum.ACT.value]
        self.form = data[self.FilingSchemaEnum.FORM.value]
        self.file_number = data[self.FilingSchemaEnum.FILE_NUMBER.value]
        self.film_number = data[self.FilingSchemaEnum.FILM_NUMBER.value]
        self.items = data[self.FilingSchemaEnum.ITEMS.value]
        self.size = data[self.FilingSchemaEnum.SIZE.value]
        self.is_xbrl = data[self.FilingSchemaEnum.IS_XBRL.value]
        self.is_inline_xbrl = data[self.FilingSchemaEnum.IS_INLINE_XBRL.value]
        self.primary_document_name = data[self.FilingSchemaEnum.PRIMARY_DOCUMENT.value]
        self.primary_document_description = data.get(self.FilingSchemaEnum.PRIMARY_DOCUMENT_DESCRIPTION.value)

        self._endpoint_format_kwargs = self.__set_endpoint_format_kwargs()
        self.primary_document_link = EndpointEnum.EDGAR_DATA_ARCHIVES.value.format(**self._endpoint_format_kwargs)

    def __set_endpoint_format_kwargs(self):
        # drop the leading zeroes from cik since they are not used in the endpoint
        cik_num = int(self.cik)
        formatted_accession_num = self.accession_number.replace("-", "")
        return {
            "CIK_NUM": cik_num,
            "ACCESSION_NUM": formatted_accession_num,
            "FILE_NAME": self.primary_document_name
        }

    def download_primary_document(self, user_agent, output_path, chunk_size=1028):
        """
        Downloads the primary document for the SEC archive
        @param user_agent: str, unique identifier needed to make request
        @param output_path: str, path to save downloaded document to
        @param chunk_size: int, number of bytes to process from request at a time
        @return: None
        """
        # TODO is there a better way of going about this that doesn't involve user supplying a user_agent?
        # Seems sort of clumsy. On paper, one request for a single document shouldn't get rate limited
        nwc = NetworkClient(user_agent)
        nwc.download_file(EndpointEnum.EDGAR_DATA_ARCHIVES, output_path, chunk_size,  **self._endpoint_format_kwargs)



from secpy.core.mixins.base_network_client_mixin import BaseNetworkClientMixin, EndpointEnum
from enum import Enum

from secpy.core.utils.cik_opts import CIKOpts


class TickerCompanyExchangeMap(BaseNetworkClientMixin):
    UNKNOWN = "UNKNOWN"
    CIK_LENGTH = 10

    def __init__(self, user_agent, **kwargs):
        """
        Handles downloading/parsing of company_tickers_exchange.json file from SEC REST API into a map of ticker -> CTEObject
        Used throughout to convert between CIK -> ticker values

        Important Note: Unlisted companies will not appear in this mapping they do not typically appear in company_tickers_exhcnage.json
        @param user_agent: Used in header of request to identify application making the request
        @param kwargs:
        """
        super().__init__(user_agent, **kwargs)
        self.__ticker_to_cte_object_mapping = None

    def lookup_ticker(self, ticker):
        """
        Search the __ticker_to_cte_object_mapping for a ticker, downloads company_tickers_exchange.json if it doesn't exist
        @param ticker: ticker to look up in map
        @return: CTEObject
        """
        self.__build_ticker_to_cte_object_mapping_if_none()
        return self.__ticker_to_cte_object_mapping[ticker]

    def lookup_cik(self, cik):
        """
        Search the __ticker_to_cte_object_mapping for a CIK, downloads company_tickers_exchange.json if it doesn't exist
        @param cik: cik to look up in map
        @return: CTEObject
        """
        self.__build_ticker_to_cte_object_mapping_if_none()
        find = [cte_object for cte_object in self.__ticker_to_cte_object_mapping.values() if cte_object.cik == cik]
        if find:
            return find[0]
        else:
            return self.UNKNOWN

    def filter_companies_by_exchange(self, exchange_enum):
        """
        Filters __ticker_to_cte_object_mapping into a mapping of only CTEObjects that are a part of the specified exchange
        @param exchange_enum: ExchangeEnum value defining what exchange to filter on
        @return: dictionary of ticker -> CTEObject where all CTEObject.exchange == exchange_enum.value
        """
        self.__build_ticker_to_cte_object_mapping_if_none()
        return {_: v for _, v in self.__ticker_to_cte_object_mapping.items() if v.exchange == exchange_enum.value}

    def ticker_to_cik_mapping(self):
        """
        Lists a mapping between ticker -> cik for each company
        @return:
        """
        self.__build_ticker_to_cte_object_mapping_if_none()
        return {_: v.cik for _, v in self.__ticker_to_cte_object_mapping.items()}

    def list_tickers(self):
        """
        Lists all tickers in the mapping
        @return: List[str]
        """
        self.__build_ticker_to_cte_object_mapping_if_none()
        return list(self.__ticker_to_cte_object_mapping.keys())

    def list_names(self):
        """
        Lists all companies in the mapping
        @return: List[str]
        """
        self.__build_ticker_to_cte_object_mapping_if_none()
        return [obj.name for obj in self.__ticker_to_cte_object_mapping.values() if obj.name != ""]

    def list_ciks(self):
        """
        Lists all ciks in the mapping
        @return: List[str]
        """
        self.__build_ticker_to_cte_object_mapping_if_none()
        return [obj.cik for obj in self.__ticker_to_cte_object_mapping.values()]

    def __build_ticker_to_cte_object_mapping_if_none(self):
        if not self.__ticker_to_cte_object_mapping:
            self.__build_ticker_to_cte_object_mapping()

    def __build_ticker_to_cte_object_mapping(self):
        response = self._validate_args_and_make_request(EndpointEnum.COMPANY_TICKER_EXCHANGE)
        response_data = response["data"]
        cte_objs = [CTEObject(obj) for obj in response_data]
        cik_to_ticker_mapping = {cte_obj.ticker: cte_obj for cte_obj in cte_objs}
        self.__ticker_to_cte_object_mapping = cik_to_ticker_mapping


class CTEObject:
    class CTESchemaEnum(Enum):
        """
        Each object in company_ticker_json is an array w/ 4 values, w/ each value representing one part of the schema
        """
        CIK = 0
        NAME = 1
        TICKER = 2
        EXCHANGE = 3

    def __init__(self, obj):
        """
        Data object class for storing data from the company_ticker_exchange.json endpoint
        CTE is short for Company Ticker Exchange.
        @param obj: dictionary to parse into CTEObject
        """
        self.cik = self.__set_cik(obj)
        self.ticker = obj[self.CTESchemaEnum.TICKER.value]
        self.exchange = obj[self.CTESchemaEnum.EXCHANGE.value]
        self.name = obj[self.CTESchemaEnum.NAME.value]

    def __set_cik(self, obj):
        cik_num = obj[self.CTESchemaEnum.CIK.value]
        return CIKOpts.format_cik(cik_num)

    def __eq__(self, other):
        if isinstance(other, CTEObject):
            return self.__dict__ == other.__dict__
        else:
            return False


class ExchangeEnum(Enum):
    """
    List of exchanges represented in the company_tickers_exchange.json file
    """
    NASDAQ = "Nasdaq"
    OTC = "OTC"
    CBOE = "CBOE"
    NYSE = "NYSE"
    NONE = ""



import re
from datetime import date


class PeriodFormatOpts:

    @staticmethod
    def validate_period_format_arg(period_format_str):
        pf_regex = re.compile("CY\d{4}(Q\dI*)*")
        match = pf_regex.fullmatch(period_format_str)
        assert match is not None, "Invalid period_format argument {}!" \
                                  "Argument should take one of the following forms:" \
                                  "CY####, CY####Q#, or CY####Q#I"
        return True

    @staticmethod
    def format_period_format_arg(period_format, use_instantaneous=False):
        instantaneous_flag = "I" if use_instantaneous else ""
        if isinstance(period_format, str):
            return period_format + instantaneous_flag
        elif isinstance(period_format, date):
            year = period_format.year
            quarter = ((period_format.month - 1) // 3) + 1
            return "CY{}Q{}{}".format(year, quarter, instantaneous_flag)
        elif isinstance(period_format, int):
            return "CY{}".format(period_format)
        else:
            raise Exception("Type of period_format must be either a string or datetime.date: {}".format(period_format))



import unittest

from secpy.core.utils.period_format_opts import PeriodFormatOpts
from datetime import date


class PeriodFormatOptsTest(unittest.TestCase):
    def test_validate_period_format_arg_on_valid_args(self):
        year = "CY2022"
        year_and_quarter = "CY2022Q1"
        year_and_quarter_instantaneous = "CY2022Q1I"
        valid_period_formats = [year, year_and_quarter, year_and_quarter_instantaneous]
        for period_format in valid_period_formats:
            self.assertTrue(PeriodFormatOpts.validate_period_format_arg(period_format))

    def test_validate_period_format_arg_on_invalid_args(self):
        cik = "0000012345"
        invalid_year = "CY22"
        invalid_quarter = "CY2022Q01"
        invalid_instantaneous_flag = "CY2022Q1F"
        invalid_period_formats = [cik, invalid_year, invalid_quarter, invalid_instantaneous_flag, invalid_instantaneous_flag]
        for invalid_period_format in invalid_period_formats:
            self.assertRaises(AssertionError, PeriodFormatOpts.validate_period_format_arg, invalid_period_format)

    def test_format_period_format_arg_str(self):
        period_format_str = "CY2022"
        actual_formatted_period = PeriodFormatOpts.format_period_format_arg(period_format_str)
        self.assertEqual(actual_formatted_period, period_format_str)

        actual_formatted_period_instantaneous = PeriodFormatOpts.format_period_format_arg(period_format_str, use_instantaneous=True)
        expected_formatted_period_instantaenous = "CY2022I"
        self.assertEqual(actual_formatted_period_instantaneous, expected_formatted_period_instantaenous)

    def test_format_period_format_arg_date(self):
        test_date_q1 = date(year=2022, month=1, day=1)
        actual_formatted_period_q1 = PeriodFormatOpts.format_period_format_arg(test_date_q1)
        expected_formatted_period_q1 = "CY2022Q1"
        self.assertEqual(actual_formatted_period_q1, expected_formatted_period_q1)

        test_date_q2 = date(year=2022, month=4, day=1)
        actual_formatted_period_q2 = PeriodFormatOpts.format_period_format_arg(test_date_q2)
        expected_formatted_period_2 = "CY2022Q2"
        self.assertEqual(actual_formatted_period_q2, expected_formatted_period_2)

        test_date_q3 = date(year=2022, month=7, day=1)
        actual_formatted_period_q3 = PeriodFormatOpts.format_period_format_arg(test_date_q3)
        expected_formatted_period_q3 = "CY2022Q3"
        self.assertEqual(actual_formatted_period_q3, expected_formatted_period_q3)

        test_date_q4 = date(year=2022, month=10, day=1)
        actual_formatted_period_q4 = PeriodFormatOpts.format_period_format_arg(test_date_q4, use_instantaneous=True)
        expected_formatted_period_q4 = "CY2022Q4I"
        self.assertEqual(actual_formatted_period_q4, expected_formatted_period_q4)

    def test_period_format_arg_int(self):
        test_year = 2022
        actual_period_format = PeriodFormatOpts.format_period_format_arg(test_year)
        expected_period_format = "CY2022"
        self.assertEqual(actual_period_format, expected_period_format)


if __name__ == '__main__':
    unittest.main()



from enum import Enum
from datetime import datetime

from secpy.core.bulk_data import BulkDataEndpoint
from secpy.core.endpoint_enum import EndpointEnum
from secpy.core.mixins.base_data_object_mixin import BaseDataObjectMixin
from secpy.core.mixins.base_endpoint_mixin import BaseEndpointMixin
from types import SimpleNamespace

from secpy.core.utils.cik_opts import CIKOpts


class CompanyFactsEndpoint(BaseEndpointMixin):
    """
    Handles the downloading and parsing of all company concepts of a given company.
    """
    _endpoint = EndpointEnum.COMPANY_FACTS

    def get_company_facts_for_ticker(self, ticker):
        """
        Gets CompanyFacts for a given ticker
        @param ticker: str, ticker to retrieve CompanyFacts for
        @return: CompanyFacts
        """
        cte_object = self._ticker_cte_map.lookup_ticker(ticker)
        return self.get_company_facts_for_cik(cte_object.cik)

    def get_company_facts_for_cik(self, cik):
        """
        Gets CompanyFacts for a given cik
        @param cik: str, cik to retrieve CompanyFacts for
        @return: CompanyFacts
        """
        response = self._validate_args_and_make_request(self._endpoint, CIK=cik)
        return CompanyFacts(response)


class CompanyFactsBulkEndpoint(BulkDataEndpoint):
    """
    Handles the downloading and parsing of the bulk CompanyFacts zip file
    """
    _endpoint = EndpointEnum.BULK_COMPANY_FACTS

    def _parse_data(self, data):
        return CompanyFacts(data)


class CompanyFacts(BaseDataObjectMixin):
    class CompanyFactsSchemaEnum(Enum):
        CIK = "cik"
        ENTITY_NAME = "entityName"
        FACTS = "facts"

    def __init__(self, data):
        """
        Container class for company facts data.
        Data is divided between one or more accounting standards (AKA taxonomies) and then is further broken down into the individual
        concepts that describe various components of reported financial data for a given company.
        Each concept contains an array of data where each element represents the value of that fact for a given filing
        @param data: dict
        """
        self.cik = self.__set_cik(data)
        self.entity_name = data.get(self.CompanyFactsSchemaEnum.ENTITY_NAME.value)
        self.taxonomies = self.__parse_taxonomies(data)

    def __set_cik(self, data):
        cik = data.get(self.CompanyFactsSchemaEnum.CIK.value, 0)
        return CIKOpts.format_cik(cik)

    def __parse_taxonomies(self, data):
        return SimpleNamespace(**{
            taxonomy_name.replace("-", "_"): SimpleNamespace(**{
                concept_name:  Concept(concept_value, concept_name) for concept_name, concept_value in taxonomy_concepts.items()
            })
            for taxonomy_name, taxonomy_concepts in data.get(self.CompanyFactsSchemaEnum.FACTS.value, {}).items()
        })

    def list_taxonomies(self):
        """
        List available taxonomies in current CompanyFacts instance
        @return: List[str]
        """
        return list(self.taxonomies.__dict__.keys())

    def get_taxonomy(self, taxonomy):
        """
        Get a particular taxonomy in the current CompanyFacts instance
        @param taxonomy: str, taxonomy to retrieve
        @return: dict of str -> Concept
        """
        return self.taxonomies.__dict__[taxonomy]

    def get_concept(self, taxonomy, concept):
        """
        Gets a particular Concept w/in a taxonomy of the current CompanyFacts instance
        @param taxonomy: str, name of taxonomy to retrieve (us-gaap, dei, etc)
        @param concept: str, name of concept to retrieve (Assets, AccountsPayableCurrent)
        @return: Concept
        """
        taxonomy_data = self.get_taxonomy(taxonomy)
        return taxonomy_data.__dict__[concept]

    def get_all_concepts(self):
        """
        Gets all concepts in every taxonomy listed under a given CompanyFacts instance
        @return: dict of Concepts
        """
        return {concept_name: concept_value for concept_dict in self.taxonomies.__dict__.values() for concept_name, concept_value in concept_dict.__dict__.items()}

    def get_statement_history(self):
        """
        Groups facts together by the form type, financial year, and financial period to form a Statement instance
        @return: List[Statement]
        """
        filing_map = {}
        for concepts in self.taxonomies.__dict__.values():
            for concept_name, concept in concepts.__dict__.items():
                form_period_unit_map = self.__get_form_period_map_for_concept(concept)
                for form_period, facts in form_period_unit_map.items():
                    if form_period not in filing_map:
                        filing_map[form_period] = Statement(facts)
                    else:
                        existing_filing_map = filing_map[form_period]
                        existing_filing_map.add_fact_to_map(concept_name, facts)
        return StatementHistory(filing_map)

    @staticmethod
    def __get_form_period_map_for_concept(concept):
        form_period_unit_map = {}
        for unit_name, facts in concept.units.__dict__.items():
            for fact in facts:
                form_period = fact.get_form_frame()
                if form_period not in form_period_unit_map:
                    form_period_unit_map[form_period] = {unit_name: fact}
                else:
                    existing_form_period_unit_mapping = form_period_unit_map[form_period]
                    existing_form_period_unit_mapping[unit_name] = fact
        return form_period_unit_map


class StatementHistory(BaseDataObjectMixin):
    def __init__(self, form_period_filing_map):
        self.__form_period_filing_map = form_period_filing_map

    def get_all_statements(self):
        """
        Gets all statements in StatementHistory
        @return: List[Statements]
        """
        return list(self.__form_period_filing_map.values())

    def get_statement_for_form_and_period(self, form, period):
        """
        Gets a Statement instance for a particular form and period if it exists
        @param form: str
        @param period: str
        @return: Statement
        """
        key = "{}_{}".format(form, period)
        return self.__form_period_filing_map[key]

    def get_statements_for_form(self, form):
        """
        Gets all statements for a particular form type
        @param form: str
        @return: List[Statement]
        """
        return [statement for key, statement in self.__form_period_filing_map.items() if key.startswith(form)]

    def get_statements_for_date_range(self, start_date=None, end_date=None, date_format="%y/%m/%d"):
        """
        Gets all statements that were filed w/in a given date range.
        If start_date but not end_date is specified, all statements filed after start_date will be returned
        If end_date but not start_date is specified, all statements filed before end_date will be returned
        @param start_date: str or datetime instance
        @param end_date: str or datetime instance
        @param date_format: str, the date format of any string start_time/end_time args
        @return: List[Statement]
        """
        assert start_date or end_date, "At least one of start_date or end_date arguments must be populated!"
        if isinstance(start_date, str):
            start_date = datetime.strptime(start_date, date_format)
        if isinstance(end_date, str):
            end_date = datetime.strptime(end_date, date_format)
        if start_date and end_date:
            assert start_date < end_date, "start_date cannot be greater than end_date!"

        result = []
        for statement in self.__form_period_filing_map.values():
            statement_filed_date = datetime.strptime(statement.filed, "%Y-%m-%d")
            if start_date and end_date and start_date < statement_filed_date < end_date:
                result.append(statement)
                continue
            if start_date and not end_date and statement_filed_date > start_date:
                result.append(statement)
            if end_date and not start_date and statement_filed_date < end_date:
                result.append(statement)
        return result


class Statement(BaseDataObjectMixin):
    def __init__(self, facts):
        """
        Contains an aggregation of all facts available for a company for a particular form type, fiscal year, and fiscal period
        @facts: Initial Unit -> Fact map instance to initialize object attributes
        """
        fact = facts[list(facts.keys())[0]]
        self.start = fact.start
        self.end = fact.end
        self.accn = fact.accn
        self.fiscal_year = fact.fiscal_year
        self.fiscal_period = fact.fiscal_period
        self.form = fact.form
        self.filed = fact.filed
        self.__facts_map = self.__initialize_facts_map(facts, fact.concept_name)

    @staticmethod
    def __initialize_facts_map(facts, concept_name):
        return {concept_name: facts}

    def add_fact_to_map(self, fact_name, unit_to_val_map):
        self.__facts_map[fact_name] = unit_to_val_map

    def get_facts_for_unit(self, fact_name, unit):
        """
        Get Fact for a given fact and unit
        @param fact_name: str, name of fact to retrieve
        @param unit: str, unit to retrieve
        @return: Fact
        """
        return self.__facts_map[fact_name][unit]

    def list_all_facts(self):
        """
        Gets list of all available Facts w/in the given statement instance
        @return: List[Fact]
        """
        return self.__facts_map.keys()

    def get_all_facts(self):
        """
        Returns all facts for the given Statement
        @return: dict, map of fact -> unit
        """
        return self.__facts_map


class HasFactMixin(BaseDataObjectMixin):
    UNITS = "units"

    def __init__(self, data, tag):
        self.tag = tag
        self.units = self.__parse_units(data)

    def __parse_units(self, data):
        return SimpleNamespace(**{
            unit_name.replace("/", "_"): [Fact(fact_data, self.tag, unit_name) for fact_data in unit_value]
            for unit_name, unit_value in data[self.UNITS].items()
        })

    def get_unit(self, key):
        """
        Gets a Fact associated w/ a given unit
        @param key: str, unit to retrieve
        @return: Fact
        """
        return self.units.__dict__[key]

    def list_units(self):
        """
        List available units
        @return: List[str]
        """
        return list(self.units.__dict__.keys())


class Concept(HasFactMixin):
    class ConceptSchemaEnum(Enum):
        LABEL = "label"
        DESCRIPTION = "description"
        UNITS = "units"

    def __init__(self, data, concept_name):
        """
        Represents a single financial concept for a particular company (Ex. AccountsPayable, AccountsReceivable, etc)
        The measurement of a concept is broken down by units (Ex. USD, USD/share, shares, etc) which are parsed into Fact instances
        @param data: dict
        """
        super().__init__(data, concept_name)
        self.label = data.get(self.ConceptSchemaEnum.LABEL.value)
        self.description = data.get(self.ConceptSchemaEnum.DESCRIPTION.value)


class Fact(BaseDataObjectMixin):
    class FactSchemaEnum(Enum):
        START = "start"
        END = "end"
        VAL = "val"
        ACCN = "accn"
        FY = "fy"
        FP = "fp"
        FORM = "form"
        FILED = "filed"
        FRAME = "frame"

    def __init__(self, data, concept_name, unit):
        """
        Represents the state of a single concept for a given company as measured by some unit at a given time for some form type
        @param data: dict
        """
        self.concept_name = concept_name
        self.unit = unit
        self.start = data.get(self.FactSchemaEnum.START.value)
        self.end = data[self.FactSchemaEnum.END.value]
        self.value = data[self.FactSchemaEnum.VAL.value]
        self.accn = data[self.FactSchemaEnum.ACCN.value]
        self.fiscal_year = data[self.FactSchemaEnum.FY.value]
        self.fiscal_period = data[self.FactSchemaEnum.FP.value]
        self.form = data[self.FactSchemaEnum.FORM.value]
        self.filed = data[self.FactSchemaEnum.FILED.value]
        self.frame = data.get(self.FactSchemaEnum.FRAME.value)

    def get_form_frame(self):
        """
        Formats a string in form_fy_fp format where form represents the form type, fy represents financial year, and
        fp represents the financial period in which the form was submitted
        @return: str
        """
        frame = self.frame or "CY{}{}".format(self.fiscal_year, self.fiscal_period)
        return "{}_{}".format(self.form, frame)



import unittest
import os

from secpy.core.bulk_data import BulkDataFileObject
from tests.testutils.mock_utils import mock_company_tickers_exchange, RESOURCES
from zipfile import ZipFile


company_tickers_exchange = mock_company_tickers_exchange()


class BulkDataFileObjectTest(unittest.TestCase):
    BULK_SUBMISSION_FILE = os.path.join(os.path.dirname(__file__),
                                        "..",
                                        "resources",
                                        "bulk_submissions.zip")

    bulk_data_archive_test_path = os.path.join(RESOURCES, "bulk_submissions.zip")

    def test_bulk_data_file_object_mappings(self):
        bulk_data_file_object = BulkDataFileObject(self.bulk_data_archive_test_path, self.company_tickers_exchange)

        expected_cik_to_filename_map = {'0000001750': 'CIK0000001750.json', '0000001800': 'CIK0000001800.json', '0000001961': 'CIK0000001961.json', '0000002034': 'CIK0000002034.json', '0000002098': 'CIK0000002098.json'}
        self.assertEqual(bulk_data_file_object.get_cik_to_filename_map(), expected_cik_to_filename_map)

        expected_ticker_to_filename_map = {'AIR': 'CIK0000001750.json', 'ABT': 'CIK0000001800.json', 'WDDD': 'CIK0000001961.json', 'ACU': 'CIK0000002098.json'}
        self.assertEqual(bulk_data_file_object.get_ticker_to_filename_map(), expected_ticker_to_filename_map)

        expected_unlisted_companies = ['0000002034']
        self.assertEqual(bulk_data_file_object.unlisted_companies, expected_unlisted_companies)

    def test_get_filelist(self):
        bulk_data_file_object = BulkDataFileObject(self.bulk_data_archive_test_path, self.company_tickers_exchange)
        actual_files = bulk_data_file_object.get_filelist()
        actual_file_names = [file.filename for file in actual_files]
        expected_files = ZipFile(self.bulk_data_archive_test_path).filelist
        expected_file_names = [file.filename for file in expected_files]

        self.assertListEqual(actual_file_names, expected_file_names)

    def test_archive_exists(self):
        bulk_data_file_object = BulkDataFileObject(self.bulk_data_archive_test_path, self.company_tickers_exchange)
        self.assertTrue(bulk_data_file_object.archive_exists)


if __name__ == '__main__':
    unittest.main()



import pathlib
from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

PACKAGE_NAME = "sec-python"
VERSION = "0.4.0-alpha"

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    author="Calvin Kessler",
    description=("Python library for interacting w/ the SEC REST API: https://www.sec.gov/edgar/sec-api-documentation"),
    license="MIT",
    install_requires=["requests==2.27.1",
                      "backoff==1.11.1",
                      "ratelimiter==1.2.0",
                      "tqdm==4.62.3"
                      ],
    keywords=["SEC", "EDGAR", "finance", "REST API wrapper"],
    url="https://github.com/McKalvan/secpy",
    download_url="https://github.com/McKalvan/secpy/archive/refs/tags/v{}.zip".format(VERSION),
    packages=find_packages(exclude="tests"),
    long_description=README,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3"
    ]
)



import backoff
import requests
from ratelimiter import RateLimiter
from tqdm import tqdm


class NetworkClient:
    __MAXIMUM_REQUESTS_PER_SEC_CAP = 10

    def __init__(self,
                 user_agent,
                 max_requests_per_sec=10,
                 max_retries=5
                 ):
        """
        Handles all requests to SEC REST API endpoints. Ensures that requests are formatted properly and are in accordance
        with the SEC's recommendations as specified here: https://www.sec.gov/privacy.htm#security
        @param user_agent: Used in header of request to identify application making the request
        @param max_requests_per_sec: Maximum number of request against the SEC REST API in one second
        @param max_retries: Maximum number of retries to make a request before giving up
        """
        self._headers = self.__set_headers(user_agent)
        self._max_requests_per_sec = self.__set_max_requests_per_sec(max_requests_per_sec)
        self.max_retries = self.__set_max_retries(max_retries)

    @staticmethod
    def __set_headers(user_agent):
        assert isinstance(user_agent, str), "user_agent arg {} is not of type string!".format(user_agent)
        assert user_agent != "", "user_agent must be a non-empty string!"
        return {
            "User-Agent": user_agent
        }

    def __set_max_requests_per_sec(self, max_requests_per_sec):
        assert isinstance(max_requests_per_sec, int) and max_requests_per_sec > 0, "max_requests_per_sec arg {} must be a positive integer!".format( max_requests_per_sec)
        assert max_requests_per_sec <= self.__MAXIMUM_REQUESTS_PER_SEC_CAP, "max_requests_per_sec {} must be less than or equal to {} or else application risks getting rate limited. " \
                                                                            "See https://www.sec.gov/privacy.htm#security for more information re. rate limiting".format(max_requests_per_sec, self.__MAXIMUM_REQUESTS_PER_SEC_CAP)
        return max_requests_per_sec

    @staticmethod
    def __set_max_retries(max_retries):
        assert isinstance(max_retries, int) and max_retries > 0, "rate_limit arg {} must be a positive integer!".format(
            max_retries)
        return max_retries

    def make_request_json(self, endpoint, **kwargs):
        """
        Makes a request to a given endpoint
        @param endpoint: EndpointEnum value
        @param kwargs: used to specify substitution variables in order to format endpoint
        @return: response in json form
        """
        return self.make_request(endpoint, **kwargs).json()

    def make_request(self, endpoint, **kwargs):
        """
        Makes a request to a given SEC REST API endpoint in the set of endpoint templates defined by EndpointEnum
        @param endpoint: EndpointEnum value
        @param kwargs: used to specify substitution variables in order to format endpoint
        @return: response
        """

        @RateLimiter(max_calls=self._max_requests_per_sec, period=1)
        @backoff.on_exception(backoff.expo,
                              requests.exceptions.RequestException,
                              max_tries=self.max_retries
                              )
        def __make_requests_helper():
            formatted_endpoint = endpoint.value.format(**kwargs)
            return requests.get(
                formatted_endpoint,
                headers=self._headers
            )

        response = __make_requests_helper()
        self.__validate_response(response)
        return response

    def download_file(self, endpoint, file_path, chunk_size, disable_progress_bar=False, **kwargs):
        """
        Downloads a file from the SEC REST API and stores it on disk
        @param disable_progress_bar: bool, disables the progress bar from being logged
        @param endpoint: EndpointEnum value
        @param file_path: output location of the file
        @param chunk_size: number of bytes to read into memory while iterating over response
        @param kwargs: used to specify substitution variables in order to format endpoint
        @return:
        """
        @RateLimiter(max_calls=self._max_requests_per_sec, period=1)
        @backoff.on_exception(backoff.expo,
                              requests.exceptions.RequestException,
                              max_tries=self.max_retries
                              )
        def __download_file_helper():
            formatted_endpoint = endpoint.value.format(**kwargs)
            response = requests.get(formatted_endpoint, stream=True, allow_redirects=True, headers=self._headers)
            content_length = int(response.headers.get('content-length'))
            initial_pos = 0
            with open(file_path, "wb") as file:
                with tqdm(total=content_length, unit_scale=True, unit="B", desc=file_path, initial=initial_pos, ascii=True, disable=disable_progress_bar) as pbar:
                    for chunk in response.iter_content(chunk_size=chunk_size):
                        file.write(chunk)
                        pbar.update(len(chunk))

        __download_file_helper()

    @staticmethod
    def __validate_response(response):
        response.raise_for_status()



from enum import Enum

from secpy.core.endpoint_enum import EndpointEnum
from secpy.core.mixins.base_data_object_mixin import BaseDataObjectMixin
from secpy.core.mixins.base_endpoint_mixin import BaseEndpointMixin
from secpy.core.utils.period_format_opts import PeriodFormatOpts


class FramesEndpoint(BaseEndpointMixin):
    """
    Handles the downloading and parsing of Frames data from the SEC REST API
    """
    _endpoint = EndpointEnum.FRAMES

    def get_frames(self, taxonomy, concept, unit, period_format, use_instantaneous=False):
        """
        Gets a Frames instance for a given taxonomy, unit, and period
        @param taxonomy: str, taxonomy to retrieve frames for (ie us-gaap, dei)
        @param concept: str, concept to retrieve frames for (ie Assets, AccountsPayableCurrent)
        @param unit: str, unit to retrieve frames for (ie USD, USD_shares
        @param period_format: str or datetime instance, time period to retrieve frames for (ie, CY2022FY, CY2021Q1)
        @param use_instantaneous: bool, use instantaneous frames
        @return: Frames instance
        """
        period_format_arg = PeriodFormatOpts.format_period_format_arg(period_format, use_instantaneous)
        response = self._validate_args_and_make_request(self._endpoint,
                                                        TAXONOMY=taxonomy,
                                                        CONCEPT=concept,
                                                        UNIT=unit,
                                                        PERIOD_FORMAT=period_format_arg
                                                        )
        return Frames(response)


class Frames(BaseDataObjectMixin):
    class FramesSchemaEnum(Enum):
        TAXONOMY = "taxonomy"
        TAG = "tag"
        CCP = "ccp"
        UOM = "uom"
        LABEL = "label"
        DESCRIPTION = "description"
        PTS = "pts"
        DATA = "data"

    def __init__(self, data):
        """
        Aggregate of CompanyFrames that represents data for a particular taxonomy/concept/unit for all available companies
        at a specified period of time.
        @param data: dict
        """
        self.taxonomy = data[self.FramesSchemaEnum.TAXONOMY.value]
        self.tag = data[self.FramesSchemaEnum.TAG.value]
        self.ccp = data[self.FramesSchemaEnum.CCP.value]
        self.uom = data[self.FramesSchemaEnum.UOM.value]
        self.label = data[self.FramesSchemaEnum.LABEL.value]
        self.description = data[self.FramesSchemaEnum.DESCRIPTION.value]
        self.pts = data[self.FramesSchemaEnum.PTS.value]
        self.data = self.__set_company_frames(data)

    def __set_company_frames(self, data):
        company_frames_arr = data[self.FramesSchemaEnum.DATA.value]
        return [CompanyFrame(obj) for obj in company_frames_arr]


class CompanyFrame(BaseDataObjectMixin):
    class CompanyFrameSchemaEnum(Enum):
        ACCN = "accn"
        CIK = "cik"
        ENTITY_NAME = "entityName"
        LOC = "loc"
        END = "end"
        VAL = "val"

    def __init__(self, data):
        """
        Represents data for a particular taxonomy/concept/unit for a single company
        @param data: dict
        """
        self.accn = data[self.CompanyFrameSchemaEnum.ACCN.value]
        self.cik = data[self.CompanyFrameSchemaEnum.CIK.value]
        self.entity_name = data[self.CompanyFrameSchemaEnum.ENTITY_NAME.value]
        self.loc = data[self.CompanyFrameSchemaEnum.LOC.value]
        self.end = data[self.CompanyFrameSchemaEnum.END.value]
        self.val = data[self.CompanyFrameSchemaEnum.VAL.value]



from enum import Enum

from secpy.company_facts import HasFactMixin
from secpy.core.endpoint_enum import EndpointEnum
from secpy.core.mixins.base_endpoint_mixin import BaseEndpointMixin


class CompanyConceptEndpoint(BaseEndpointMixin):
    """
    Handles the downloading and parsing of individual concepts for a given company
    """
    _endpoint = EndpointEnum.COMPANY_CONCEPT

    def get_company_concept_for_ticker(self, ticker, taxonomy, concept):
        """
        Gets CompanyConcept for a given ticker, taxonomy, and concept
        @param ticker: str, ticker to retrieve
        @param taxonomy: str, taxonomy to retrieve (ie us-gaap, dei)
        @param concept: str, concept to retrieve (ie Assets, AccountsPayableCurrent)
        @return: CompanyConcept
        """
        cte_object = self._ticker_cte_map.lookup_ticker(ticker)
        return self.get_company_concept_for_cik(cte_object.cik, taxonomy, concept)

    def get_company_concept_for_cik(self, cik, taxonomy, concept):
        """
        Gets CompanyConcept for a given cik, taxonomy, and concept
        @param cik: str, cik to retrieve
        @param taxonomy: str, taxonomy to retrieve (ie us-gaap, dei)
        @param concept: str, concept to retrieve (ie Assets, AccountsPayableCurrent)
        @return: CompanyConcepts
        """
        response = self._validate_args_and_make_request(self._endpoint, CIK=cik, TAXONOMY=taxonomy, CONCEPT=concept)
        return CompanyConcept(response)


class CompanyConcept(HasFactMixin):
    class CompanyConceptSchema(Enum):
        CIK = "cik"
        TAXONOMY = "taxonomy"
        TAG = "tag"
        LABEL = "label"
        DESCRIPTION = "description"
        ENTITY_NAME = "entityName"
        UNITS = "units"

    def __init__(self, data):
        """
        Container class for CompanyConcept data
        @param data: dict
        """
        super().__init__(data, data[self.CompanyConceptSchema.TAG.value])
        self.cik = data[self.CompanyConceptSchema.CIK.value]
        self.taxonomy = data[self.CompanyConceptSchema.TAXONOMY.value]
        self.label = data.get(self.CompanyConceptSchema.LABEL.value)
        self.description = data.get(self.CompanyConceptSchema.DESCRIPTION.value)
        self.entity_name = data.get(self.CompanyConceptSchema.ENTITY_NAME.value)



import logging

from secpy.core.mixins.base_network_client_mixin import BaseNetworkClientMixin
from secpy.core.ticker_company_exchange_map import TickerCompanyExchangeMap


class BaseEndpointMixin(BaseNetworkClientMixin):
    _endpoint = None

    def __init__(self, user_agent, **kwargs):
        """
        Base class to be inherited by all classes that interact directly w/ the SEC REST API
        @param user_agent: unique identifiers to use in headers when making requests to SEC REST API
        @param kwargs: Misc
        """
        super().__init__(user_agent, **kwargs)
        self._logger = self.__set_logger()
        self._ticker_cte_map = TickerCompanyExchangeMap(user_agent, **kwargs)

    @staticmethod
    def __set_logger():
        logging.basicConfig(format='%(asctime)s- %(pathname)s - %(levelname)s - %(message)s')
        return logging.getLogger("secpy")



import json


class BaseDataObjectMixin:
    def to_json(self):
        """
        Converts object to JSON dict
        @return: dict
        """
        return json.dumps(self, default=lambda o: o.__dict__)
